# -*- coding: utf-8 -*-

"""
pyDGSA sub-package that provides clustering using KMedoids.

Credit for this sub-package goes to:
David Zhen Yin, Aug 2019

Updated: 
Zach Perzan, June 2020
Zach Perzan, Mar 2023
"""

import numpy as np
import matplotlib.pyplot as plt

def _get_init_centers(n_clusters, n_samples):
    '''return random points as initial centers'''
    init_ids = []
    while len(init_ids) < n_clusters:
        _ = np.random.randint(0,n_samples)
        if not _ in init_ids:
            init_ids.append(_)
    return init_ids
	
def _get_cost(dist_meds, currentMedoids):
    '''return total cost and cost of each cluster
    -----
    currentMedoids: the current Medoids
    dist_meds: paired distances between all data points and each Medoid
    '''
    costs = np.zeros(len(currentMedoids))
    dis_min=np.min(dist_meds,axis=0)
    for i in range(len(currentMedoids)):
        clst_mem_ids = np.where(dist_meds[i]==dis_min)[0]
        costs[i] = np.sum(dist_meds[i][clst_mem_ids])
    return np.sum(costs)

def _kmedoids_run(X, n_clusters, max_iter=3000, tol=1e-4):
    '''
    Main function for runing the k-medoids clustering
    -------------
    X: the squareform distance array for all samples, (#samples, #samples)
    n_cluster: number of clusters
    max_iter: maximum number of clusters
    tol: the tolerance to stop the iterations, in percentage; 
                i.e.: if tol=0.01, it means if the cost function decrease is less than 1%, the iteraction will stop.
    '''
    
    n_samples = X.shape[0]

    ''' Initialize the medoids'''
    currentMedoids = np.asarray(_get_init_centers(n_clusters,n_samples))
    
    '''Calcualte the total cost of the initial medoids'''
    costs_iters=[]
    dist_meds = X[currentMedoids]
    tot_cos = _get_cost(dist_meds,currentMedoids)
    costs_iters.append(tot_cos)
    cc = 0
    
    for i in range(max_iter):
        dist_meds = X[currentMedoids]
        '''Associate  each data point to the closest medoid
            And calcualte the total cost'''
        tot_cos = _get_cost(dist_meds,currentMedoids)
        '''Get new mediods o'''
        newMedoids = []
        while len(newMedoids) == 0:
            for j in range(n_clusters):
                o = np.random.choice(n_samples)
                if (o not in currentMedoids and o not in newMedoids):
                    newMedoids.append(o)
        newMedoids = np.asarray(newMedoids).astype(int)
        dist_meds_ = X[newMedoids]

        tot_cos_ = _get_cost(dist_meds_, newMedoids)
        '''Swap newmediods with the current mediod if cost decreases'''
        if (tot_cos_ - tot_cos)<0:
            currentMedoids = newMedoids
            costs_iters.append(tot_cos_)
            cc =+1
            if abs(costs_iters[cc]/costs_iters[cc-1]-1)<tol:
                '''Associated  data points to the final calucated medoids (reached by tolerance)'''
                clsts_membr_ids = []
                dis_min=np.min(dist_meds,axis=0)
                for k in range(n_clusters):
                    clst_mem_ids = np.where(dist_meds[k]==dis_min)[0]
                    clsts_membr_ids.append(clst_mem_ids)

                return currentMedoids, clsts_membr_ids, costs_iters
                break
            
    costs_iters = np.asarray(costs_iters)
    '''Associated  data points to the final calucated medoids (reached by maximum iters)'''
    clsts_membr_ids = []
    dist_meds = X[currentMedoids]
    dis_min=np.min(dist_meds,axis=0)
    for k in range(n_clusters):
        clst_mem_ids = np.where(dist_meds[k]==dis_min)[0]
        clsts_membr_ids.append(clst_mem_ids)

    return currentMedoids, clsts_membr_ids, costs_iters

class KMedoids(object):
    '''
    Main API of KMedoids Clustering
    Parameters
    --------
        X: the input ndarray data for k-medoids clustering, (#samples, #features)
        n_clusters: number of clusters
        max_iter: maximum number of iterations
        tol:  the tolerance to stop the iterations, in percentage; 
                    i.e.: if tol=0.01, it means if the cost function decrease is less than 1%, the iteraction will stop.
    Attributes
    --------

        Medoids   :  cluster Medoids id
        costs_itr   :  array of costs for each effective iterations
        clst_membr_ids   :  each cluster members' sample-ids in the input ndarray X 
    
    Methods
    -------
        model = KMedoids(n_clusters = #, max_iter=#<opertional>, tol==#<opertional>)
        Medoids, cluster_id = model.fit(X, plotit=True/False): fit the model, 
                                                    it returns the center medoids id in the ndarray X, 
                                                    and  each cluster members' sample-ids in X.
    '''
    def __init__(self, n_clusters, max_iter=10000, tol=0.005):
        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.tol = tol

    def fit_predict(self, X):
        '''Run the main k-medoids function'''
        self.cluster_centers_, clst_membr_ids, self.costs_itr = _=_kmedoids_run(X, self.n_clusters, self.max_iter,self.tol)
        
        # Convert clst_membr_ids to labels
        labels = np.ones(X.shape[0], dtype='int32')
        for i in range(self.n_clusters):
            labels[clst_membr_ids[i]] = i

        return labels, self.cluster_centers_



# -*- coding: utf-8 -*-

"""
pyDGSA sub-package that provides helper functions used in the 
pyDGSA.dgsa.dgsa_interactions submodule

This file includes:
interact_boot_distances
interact_distance
"""

import numpy as np
from tqdm.notebook import tqdm


def interact_distance(cond_idx, parameters, clusters, thresholds, percentiles):
    """For a single conditional parameter, calculate the L1-norm distance between 
    the class-conditional distributions of all other parameters and the class-
    conditional distributions additionally conditioned on another parameter. 
    
    params:
        cond_idx [int]: index of the conditional parameter within the columns 
                of the parameters array
        parameters [array]: array of shape (n_simulations, n_parameters) 
                containing parameters used to generate each simulation
        clusters [dict(array)]: dict of len(n_clusters). Values are the indices 
                within each cluster, corresponding to the rows of parameters
        thresholds [array]: array of shape (n_bins-1, n_parameters) containing the 
                thresholds at which to separate each conditional parameter
        percentiles [array]: array of percentiles at which to evaluate cdf
    
    returns:
        cdf_distances [array]: array of shape (n_parameters-1, n_clusters, 
                n_bins) containing the L1-norm distances for each 
                parameter/cluster/bin combination
    """

    # Get number of params, bins and clusters
    n_parameters = parameters.shape[1]
    n_bins = thresholds.shape[0] + 1
    n_clusters = len(clusters)

    cdf_distances = np.zeros((n_parameters, n_clusters, n_bins), dtype='float64')

    # Loop through cluster and bin
    for nc in range(n_clusters):
        # Indices comprising this cluster
        c_idx = clusters[nc] 
        
        # Priors for all parameters, ie F(p|C)
        q_prior = np.percentile(parameters[c_idx, :], percentiles, axis=0) 

        for nb in range(n_bins): 
            # For each bin, first calc b_mask -- mask of params within each bin    
            # The first bin
            if nb == 0: 
                threshold = thresholds[nb, cond_idx]
                b_mask = parameters[:, cond_idx] <= threshold
            # The last bin
            elif nb == (n_bins - 1): 
                threshold = thresholds[-1, cond_idx]
                b_mask = parameters[:, cond_idx] > threshold
            # Middle bins
            else: 
                low_thresh = thresholds[nb - 1, cond_idx]
                high_thresh = thresholds[nb, cond_idx]
                b_mask = (parameters[:, cond_idx] > low_thresh) & \
                         (parameters[:, cond_idx] <= high_thresh)

            # Indices comprising this bin
            b_idx = np.argwhere(b_mask) 
            
            # Indices within this cluster AND bin
            bc_idx = np.intersect1d(c_idx, b_idx.flatten(), assume_unique=True) 

            # If no params within a cluster/bin, distance = nan
            if len(bc_idx) == 0:
                cdf_distances[:, nc, nb] = np.nan
            else:
                q_inter = np.percentile(parameters[bc_idx, :], percentiles, axis=0) 
                cdf_distances[:, nc, nb] = np.sum(abs(q_inter - q_prior), axis=0)

    # Delete row of zeros (parameter conditioned on itself) from array
    cdf_distances = np.delete(cdf_distances, cond_idx, axis=0)
                    
    return cdf_distances


def interact_boot_distance(cond_idx, parameters, clusters, thresholds, 
                           percentiles, n_boots=3000, alpha=0.95):
    """For a single parameter, performs a resampling procedure that calculates
    the alpha-quantile of the L1-norm for each parameter given the conditional 
    parameter, each cluster and each bin.
    
    params:
        cond_idx [int]: index of the conditional parameter within the columns 
                of the parameters array
        parameters [array]: array of shape (n_simulations, n_parameters) 
                containing parameters used to generate each simulation
        clusters [dict(array)]: dict of len(n_clusters). Values are the indices 
                within each cluster, corresponding to the rows of parameters
        thresholds [array]: array of shape (n_bins-1, n_parameters) containing the 
                thresholds at which to separate each conditional parameter
        percentiles [array]: array of percentiles at which to evaluate cdf
        n_boots [int]: number of iterations to perform resampling. Default: 3000
        alpha [float]: alpha-quantile at which to evaluate L1-norm. Can be 
                either [0, 1] (quantile) or [0, 100] (percentile)
    
    returns:
        boot_distances [array]: array of shape (n_parameters-1, n_clusters, 
                n_bins) containing the alpha-quantile L1-norm distances for 
                each parameter/cluster/bin combination
    """

    # Check alpha input and scale as needed
    if (alpha >= 0) & (alpha <= 1):
        pass
    elif (alpha < 0):
        raise ValueError("alpha-percentile must be greater than zero")
    elif (alpha > 100):
        raise ValueError("alpha-percentile must be between 0 and 100 inclusive")
    else:
        alpha = alpha / 100
    
    # Get number of params, bins and clusters
    n_parameters = parameters.shape[1]
    n_bins = thresholds.shape[0] + 1
    n_clusters = len(clusters)

    boot_distances = np.zeros((n_parameters, n_clusters, n_bins), dtype='float64')

    # Loop through each parameter, cluster and bin
    for nc in tqdm(range(n_clusters), desc='Resampling parameter %d' % cond_idx, leave=False):
        # Indices comprising this cluster
        c_idx = clusters[nc] 
        
        # Priors for all parameters, ie F(p|C)
        q_prior = np.percentile(parameters[c_idx, :], percentiles, axis=0) 

        for nb in range(n_bins): 
            # The first bin
            if nb == 0: 
                threshold = thresholds[nb, cond_idx]
                b_mask = parameters[:, cond_idx] <= threshold
            # The last bin
            elif nb == (n_bins - 1): 
                threshold = thresholds[-1, cond_idx]
                b_mask = parameters[:, cond_idx] > threshold
            # Middle bins
            else: 
                low_thresh = thresholds[nb - 1, cond_idx]
                high_thresh = thresholds[nb, cond_idx]
                b_mask = (parameters[:, cond_idx] > low_thresh) & (parameters[:, cond_idx] <= high_thresh)

            b_idx = np.argwhere(b_mask)
            bc_idx = np.intersect1d(c_idx, b_idx.flatten(), assume_unique=True)
            
            # If fewer than 4 values in this bin, set to nan
            if len(bc_idx) <= 3:
                boot_distances[:, nc, nb] = np.nan
            else:
                boot_dist = np.zeros((n_boots, n_parameters))
                for n_boot in range(n_boots):
                    boot_idx = np.random.choice(c_idx, len(bc_idx), replace=False)
                    boot_inter = np.percentile(parameters[boot_idx, :], percentiles, axis=0) # cdf of cond_idx conditioned to n_p in bin nb
                    boot_dist[n_boot] = np.sum(abs(boot_inter - q_prior), axis=0)
                
                boot_dist[:, cond_idx] = np.nan
                boot_distances[:, nc, nb] = np.quantile(boot_dist, alpha, axis=0)

    # Delete row of zeros (parameter conditioned on itself) from array
    boot_distances = np.delete(boot_distances, cond_idx, axis=0)

    return boot_distances





# -*- coding: utf-8 -*-

"""
pyDGSA sub-package that provides useful utilities for plotting dgsa results

This file includes:
reformat_interactions
vert_pareto_plot
plot_cdf
calc_bubble_distances
bubble_plot
interaction_matrix
"""


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import matplotlib
from sklearn.manifold import MDS
from scipy.stats import gaussian_kde

def reformat_interactions(df):
    """Given a dataframe with columns 'k1 | k2', output a dataframe with
    main parameters as the columns and conditional parameters as the rows."""
    
    for i, interaction in enumerate(df.index.tolist()):
        main_param = interaction.split('|')[0].strip()
        cond_param = interaction.split('|')[1].strip()

        if i == 0:
            int_df = pd.DataFrame(index=[main_param], columns=[cond_param], 
                                  dtype='float64')
        int_df.loc[cond_param, main_param] = df.loc[interaction, 'sensitivity']
       
    # Sort df before returning it
    int_df.sort_index(axis=0, inplace=True)
    int_df.sort_index(axis=1, inplace=True)
    
    return int_df


def vert_pareto_plot(df, np_plot='+5', fmt=None, colors=None, confidence=False,
                     figsize=None):
    """Generate a vertical Pareto plot of sensitivity analysis results.
    
    params:
        df [DataFrame]: pandas dataframe containing the sensitivity analysis
                results. If fmt == 'max' or 'mean', df contains a single column
                with indices correspond to the parameters. If fmt == 'cluster_avg', 
                the columns of df correspond to the number of clusters and 
                the rows correspond to the parameters.
        np_plot [str|int]: number of parameters to plot. Default: '+5'
                -'all': plot all parameters
                -n: plot n parameters, where n is an int
                -'+n': plot all parameters with sensitivity >1, plus 
                       the n next most sensitive parameters
        fmt [str]: format of df. Optional, will interpret fmt based on df shape
                -'mean': single sensitivity was passed per param, therefore
                         display a single bar per param
                -'max': same as 'mean'
                -'cluster_avg': cluster-specific sensitivities were passed, so
                            display cluster's sensitivity separately
                -'bin_avg': bin-specific sensitivities were passed, so display
                            each bin's sensitivity separately
                -'indiv': plots the sensitivity for each bin/cluster combination
                        separately
        confidence [bool]: whether or not to plot confidence bars. Default is False, 
                but must be included in df if confidence == True.
        colors [list(str|int)]: list of clusters colors to use when plotting, 
                either specified as rgba tuples or strings of matplotlib named 
                colors. Only used when fmt='cluster_avg' or 'indiv'
   
    returns:
        fig: matplotlib figure handle
        ax: matplotlib axis handle
    """
    
    # Total np (number of parameters)
    np_total = df.shape[0] 
    
    # Figure out fmt if not explicitly provided
    if fmt is None:
        if isinstance(df.columns, pd.MultiIndex):
            fmt = 'indiv'
        elif df.shape[1] > 1 and 'confidence' not in df.columns:
            # Could be either 'cluster_avg' or 'bin_avg'
            if 'Cluster' in df.columns[0]:
                fmt = 'cluster_avg'
            elif 'Bin' in df.columns[0]:
                fmt = 'bin_avg'
            else:
                raise ValueError("Could not determine fmt. Please pass explicitly.")
        else:
            # Note that 'mean' also includes 'max' format from the analysis
            fmt = 'mean'
            if 'confidence' in df.columns:
                cdf = df['confidence'].copy()
                # copy to avoid altering input df
                df = df.drop('confidence', axis=1).copy() 
    
    if fmt == 'indiv' and colors is not None:
        if isinstance(colors[0], str):
            # Convert named colors to rgba
            named_colors = colors
            colors=[]
            for color in named_colors:
                colors.append(matplotlib.colors.to_rgba(color))
        
    if fmt == 'cluster_avg':
        # Check if confidence bounds were provided by counting 
        # columns that end with "_conf"
        conf_cols = [col for col in df.columns if col[-5:] == '_conf']
        if len(conf_cols) > 0:
            cdf = df[conf_cols].copy()
            df = df.drop(conf_cols, axis=1).copy()
            
    # Get number of parameters with sensitivity >= 1
    np_sensitive = np.any((df.fillna(0).values >= 1), axis=1).sum()

    # Figure out how many parameters to plot
    if isinstance(np_plot, str):
        if np_plot == 'all':
            np_max_plot = np_total
        elif np_plot[0] == '+':
            np_max_plot = np_sensitive + int(np_plot[1:])
        else:
            raise ValueError("np_plot must be 'all', 'n', or '+n', where n is an int")
    elif isinstance(np_plot, int):
        np_max_plot = np_plot

    # Ensure that requested # of params to plot is not larger than total # of params
    if np_max_plot > np_total:
        np_max_plot = np_total

    # Y-position of bars
    y_pos = np.arange(np_max_plot)
    
    if fmt == 'mean' or fmt == 'max':
        # Sort so most sensitive params are on top
        df.sort_values(by=df.columns[0], ascending=False, inplace=True)
        data = df.values[:np_max_plot, :].squeeze()
        params = df.index.tolist() # Get list of params after sorting
        
        yticks = y_pos 

        # Error bars (confidence interval); by default these are plotted, but of 
        # length 0 if confidence == False
        if confidence:
            xerr = cdf[df.index[:np_max_plot]].values/2
        else:
            xerr = 0

        # Values are color-coded. If confidence intervals are provided
        if confidence:
            colors = np.asarray([[1, 1, 1, 0.8]]*np_max_plot)
            # > confidence interval red
            colors[data - xerr > 1] = [1, 0, 0, 0.8]
            # < confidence interval blue
            colors[data + xerr < 1] = [0, 0, 1, 0.8]
        else:
            if np_sensitive > 0:
                colors = np.asarray([[1, 0, 0, 0.8]]*np_max_plot)
            else:
                colors = np.asarray([[0, 0, 1, 0.8]]*np_max_plot)

            if np_max_plot > np_sensitive:
                if np_sensitive > 0:
                    colors[np_sensitive] = [1, 1, 1, 1]
                colors[np_sensitive+1:] = [0, 0, 1, 0.8] 

        if figsize is None:
            fig_height = int(np_max_plot/2)
            figsize = (5, fig_height)

        # Create figure and add barh
        fig, ax = plt.subplots(figsize=figsize)
        ax.barh(y_pos, data, color=colors, edgecolor='k', xerr=xerr)

    elif fmt == 'cluster_avg':
        n_clusters = df.shape[1]
        
        # Sort by mean sensitivity across clusters 
        sort_df = df.mean(axis=1).sort_values(ascending=False)
        df = df.reindex(sort_df.index)
        params = df.index.tolist() # Get list of params after sorting
        
        # Add error bars if confidence=True, otherwise set length to 0
        if confidence:
            xerr = cdf.loc[df.index, :].values
        else:
            xerr = df.loc[df.index, :].values*0
        
        height = 1/(n_clusters+1)
        yticks = y_pos - (height*(n_clusters-1)/2)
        
        if colors is None:
            colors = []
            cmap = matplotlib.cm.get_cmap('Set1')
            for i in range(n_clusters):
                colors.append(cmap(i))

        # Create figure
        if figsize is None:
            fig_height = int(np_max_plot/2*1.5)
            figsize = (5, fig_height)
            
        fig, ax = plt.subplots(figsize=figsize)

        # Add bars for each cluster
        for i in range(n_clusters):
            ax.barh(y_pos - height*i, df.iloc[:np_max_plot, i], height=height, 
                    color=colors[i], edgecolor='k', label=df.columns.tolist()[i],
                    xerr=xerr[:np_max_plot, i])
        ax.legend()
        
    elif fmt == 'bin_avg':
        n_bins = df.shape[1]
        
        # Sort by mean sensitivity across bins 
        sort_df = df.mean(axis=1).sort_values(ascending=False)
        df = df.reindex(sort_df.index)
        params = df.index.tolist()

        yticks = y_pos

        if colors is None:
            cmap = matplotlib.cm.get_cmap('Set1')
            colors = cmap(1)

        # Create color array by decreasing alpha channel for each bin
        color_array = np.tile(colors, (n_bins, 1))
        for i in range(n_bins):
            color_array[i, 3] = (i+1)/(n_bins)

        if figsize is None:
            fig_height = int(np_max_plot/2*1.5)
            figsize = (5, fig_height)
            
        fig, ax = plt.subplots(figsize=figsize)

        for i in range(n_bins):
            width = df.iloc[:np_max_plot, i]
            left = df.iloc[:np_max_plot, :i].sum(axis=1)
            b = ax.barh(y_pos, width=width, left=left, color=color_array[i], 
                        edgecolor='k')
                
            # Increase linewidth for parameters that are sensitive
            for w in enumerate(width.tolist()):
                if w[1] > 1:
                    b[w[0]].set_linewidth(2.5)
        
    elif fmt == 'indiv':
        n_clusters, n_bins = df.columns.levshape
        
        # Split df into sensitive and non-sensitive df's, sort each, then re-combine
        # Can't just sort on mean sensitivity, otherwise a sensitive parameter
        # could get left out because its mean might not be within the top 
        # np_max_plot most sensitive, even though a single bin is >= 1
        mask = np.any((df.fillna(0).values >= 1), axis=1)
        sens = df[mask].copy()
        nsens = df[~mask].copy()
        sort_sens = sens.mean(axis=1).sort_values(ascending=False)
        sens = sens.reindex(sort_sens.index)
        sort_nsens = nsens.mean(axis=1).sort_values(ascending=False)
        nsens = nsens.reindex(sort_nsens.index)
        df = sens.append(nsens)

        params = df.index.tolist()
        df.fillna(0, inplace=True)
        height = 1/(n_clusters+1)
        yticks = y_pos - (height*(n_clusters-1)/2)

        if colors is None:
            cmap = matplotlib.cm.get_cmap('Set1')
            colors = []
            for i in range(n_clusters):
                colors.append(cmap(i))

        idx = pd.IndexSlice
                
        # Create color array by decreasing alpha channel for each bin
        color_array = np.zeros((n_clusters, n_bins, 4), dtype='float64')
        for i in range(n_clusters):
            for j in range(n_bins):
                color_array[i, j, :] = colors[i]
                color_array[i, j, 3] = (j+1)/(n_bins)

        fig_height = int(np_max_plot/2*1.5)
        fig, ax = plt.subplots(figsize=(5, fig_height))

        for i in range(n_clusters):
            for j in range(n_bins):
                col_idx = i*n_bins + j
                width = df.iloc[:np_max_plot, col_idx]
                left = df.iloc[:np_max_plot, n_bins*i:col_idx+1].sum(axis=1) - width
                y = y_pos - height*i
                if j == n_bins - 1:
                    # Add label to last bin
                    b = ax.barh(y, width=width, height=height, left=left, 
                                color=color_array[i, j], edgecolor='k', 
                                label=df.columns.tolist()[i*n_bins][0])
                else:
                    b = ax.barh(y, width=width, height=height, left=left, 
                                color=color_array[i, j], edgecolor='k')
                    
                # Increase linewidth for parameters that are sensitive
                for w in enumerate(width.tolist()):
                    if w[1] > 1:
                        b[w[0]].set_linewidth(2.5)
        
        leg = ax.legend()
        # Ensure that linewidths in the legend are all 1.0 pt
        for legobj in leg.legendHandles:
            legobj.set_linewidth(1)

    # Add vertical line and tick labels
    if fmt not in ['indiv', 'bin_avg']:
        ax.axvline(1, color='k', linestyle='--')
    ax.set(yticks=yticks, yticklabels=params[:np_max_plot], xlabel='Sensitivity')
    ax.invert_yaxis()
    
    # Move xaxis label and ticks to top
    ax.xaxis.tick_top()
    ax.xaxis.set_label_position('top')
        
    return fig, ax

def get_param_idx(parameter, parameter_names, n_parameters):
    """Regardless of format given, get the index of a specific parameter,
    with error checking."""
    
    # If parameter is given as string, look for it in parameter_names
    if isinstance(parameter, str):
        if parameter_names is None:
            raise ValueError('Must provide list of parameter names if providing \
                parameter to plot as a string.')
        elif parameter not in parameter_names:
            raise ValueError('Could not find %s in parameter_names' % parameter)
        else:
            param_idx = parameter_names.index(parameter)
            
    elif isinstance(parameter, int):
        if parameter > (n_parameters - 1):
            raise ValueError('Parameter index passed (%s) is greater than the \
                             length of parameter_names' % parameter)
        else:
            param_idx = parameter
            
    return param_idx


def plot_interact_cdf(parameters, labels, main_param, cond_param, cluster_names=None, 
                      colors=None, figsize=(12,5), parameter_names=None, n_bins=3,
                      bin_labels=None, plot_prior=True):
 
    # Check input
    n_clusters = labels.max() + 1
    n_parameters = parameters.shape[1]

    if colors is None:
        colors = []
        cmap = matplotlib.cm.get_cmap('Set1')
        for i in range(n_clusters):
            colors.append(cmap(i))

    # Create color array by decreasing alpha channel for each bin
    color_array = np.zeros((n_clusters, n_bins, 4), dtype='float64')
    for i in range(n_clusters):
        for j in range(n_bins):
            color_array[i, j, :] = colors[i]
            color_array[i, j, 3] = (j+1)/(n_bins)

    if cluster_names is None:
        cluster_names = ['Cluster %s' %i for i in range(n_clusters)]

    if bin_labels is None:
        bin_labels = ['Bin %s' %i for i in n_bins]

    # Calculate the thresholds separating bins for each parameter
    # thresholds is of shape (n_bins, n_parameters)
    thresholds = np.quantile(parameters, [b/n_bins for b in range(1, n_bins)], 
                             axis=0)

    # Get the indices of the main and cond params
    mpidx = get_param_idx(main_param, parameter_names, n_parameters)
    cpidx = get_param_idx(cond_param, parameter_names, n_parameters)

    if parameter_names is None:
        parameter_names = ['Parameter %s' %i for i in range(n_parameters)]

    percentiles = np.arange(1, 100)

    fig, ax = plt.subplots(1, n_clusters, figsize=figsize, tight_layout=True, sharey=True)

    for nc in range(n_clusters):
        c_idx = np.where(labels==nc)
        q_prior = np.percentile(parameters[c_idx, mpidx], percentiles)
        if plot_prior:
            ax[nc].plot(np.percentile(parameters[:,mpidx], percentiles), 
                        percentiles, color='k', linestyle='--', label='$F$ ($X_i$)')
        ax[nc].plot(q_prior, percentiles, color=color_array[nc, -2], linestyle=':', 
                    label='$F$ ($X_i$ | %s)' % cluster_names[nc])

        for nb in range(n_bins): 
            # For each bin, first calc b_mask -- mask of params within each bin    
            # The first bin
            if nb == 0: 
                threshold = thresholds[nb, cpidx]
                b_mask = parameters[:, cpidx] <= threshold
            # The last bin
            elif nb == (n_bins - 1): 
                threshold = thresholds[-1, cpidx]
                b_mask = parameters[:, cpidx] > threshold
            # Middle bins
            else: 
                low_thresh = thresholds[nb - 1, cpidx]
                high_thresh = thresholds[nb, cpidx]
                b_mask = (parameters[:, cpidx] > low_thresh) & \
                         (parameters[:, cpidx] <= high_thresh)

            # Indices comprising this bin
            b_idx = np.argwhere(b_mask) 

            # Indices within this cluster AND bin
            bc_idx = np.intersect1d(c_idx, b_idx.flatten(), assume_unique=True) 
            if len(bc_idx) > 2:
                q_inter = np.percentile(parameters[bc_idx, mpidx], percentiles, axis=0)
                label = '$F$ ($X_i$ | %s, %s)' % (cluster_names[nc], bin_labels[nb])
                ax[nc].plot(q_inter, percentiles, label=label, 
                            color=color_array[nc, nb])

        ax[nc].set(xlim=[parameters[:, mpidx].min(), parameters[:, mpidx].max()],
                  xlabel=parameter_names[mpidx], title=cluster_names[nc])
        ax[nc].legend()
    ax[0].set(ylabel='% of simulations')

    return fig, ax


def calc_bubble_distances(int_df):
    """Given a dataframe of asymmetric interactions, calculate 
    the symmetric distance for a bubble plot, as proposed by Park 
    et al. (2016)"""

    ref_df = reformat_interactions(int_df)
    sym_df = ref_df.copy()

    # Drop main_params that are not in cond_params
    for main_param in ref_df.columns:
        if main_param not in ref_df.index:
            sym_df.drop(main_param, axis=1, inplace=True)


    for cond_param in sym_df.index:
        for main_param in sym_df.columns:
            if main_param == cond_param:
                sym_df.loc[cond_param, main_param] = 0
            else:
                int1 = ref_df.loc[cond_param, main_param]
                int2 = ref_df.loc[main_param, cond_param]
                sym_df.loc[cond_param, main_param] = 2/(int1 + int2)
                
    return sym_df
         
         
def bubble_plot(main_df, int_df, figsize=(5,5)):
    
    sym_df = calc_bubble_distances(int_df)
    
    # Calculate MDS coordinates of the 
    mds = MDS(n_components=2, dissimilarity='precomputed', random_state=0)
    mds_dist = mds.fit_transform(sym_df.values)

    fig, ax = plt.subplots(figsize=figsize, tight_layout=True)

    x = mds_dist[:, 0]
    y = mds_dist[:, 1]
    sizes = main_df.loc[sym_df.index, 'sensitivity']*1000
    colors = []
    for sens in main_df.loc[sym_df.index, 'sensitivity']:
        if sens > 1:
            colors.append('red')
        else:
            colors.append('blue')

    sc = ax.scatter(x, y, s=sizes, c=colors, alpha=0.5)

    for i, label in enumerate(sym_df.index):
        ax.annotate(label, (x[i], y[i]))

    # Remove tick labels
    ax.set(yticklabels=[], xticklabels=[])

    return fig, ax


def interaction_matrix(df, figsize=(5,5), fontsize=12, nan_color='gray'):
    """Given a dataframe of asymmetric sensitivities, plot a color-coded 
    matrix of those sensitivities.
    
    params:
        df [DataFrame]: pandas dataframe containing the sensitivity analysis
                results. 
        figsize [tuple(float)]: matplotlib figure size in inches. Optional, 
                default: (5,5)
        fontsize [tuple(float)]: font size for the text values in the matrix. 
                Optional, default: 12
        nan_color [str]: a matplotlib named color for displaying NaN values
        
    returns:
        fig: matplotlib figure handle
        ax: matplotlib axis handle
    """
    int_df = reformat_interactions(df)
    main_params = int_df.columns.tolist()
    cond_params = int_df.index.tolist()

    # Separate out sensitive and insensitive parameters
    # and plot using separate color ramps
    mask = int_df.values > 1

    sens = int_df.values.copy()
    insens = int_df.values.copy()
    sens[~mask] = np.nan
    insens[mask] = np.nan

    # Create separate array of NaN's as nan_color
    nan_rgb = matplotlib.colors.to_rgba(nan_color) # convert str to rgb values
    nancells = np.tile(int_df.values.copy(), (4,1,1))

    # Swap all NaN's to RGBA and everything else to 0
    nancells[~np.isnan(nancells)] = 0
    for i, cvalue in enumerate(nan_rgb):
        nancells[i, np.isnan(nancells[i, :, :])] = cvalue

    fig, ax = plt.subplots(figsize=figsize)

    ax.imshow(insens, vmin=0, vmax=1.3, 
              cmap='Blues', interpolation='none')
    if not np.all(np.isnan(sens)):
        ax.imshow(sens, vmin=0.7, vmax=np.nanmax(sens)*1.5, 
                  cmap='Reds', interpolation='none')
    ax.imshow(nancells.swapaxes(0,2).swapaxes(0,1))

    ax.set(xticks=range(len(main_params)), yticks=range(len(cond_params)), 
           yticklabels=cond_params)
    
    # Add rotated xtick labels
    ax.set_xticklabels(labels=main_params, rotation=90)

    for i, main_param in enumerate(main_params):
        for j, cond_param in enumerate(cond_params):
            value = int_df.loc[cond_param, main_param]
            if not np.isnan(value):
                text = '%0.2f' % value
                ax.text(i, j, text, fontsize=fontsize,
                        horizontalalignment='center', verticalalignment='center')
    
    # Label axes
    ax.set(xlabel='Main parameter', ylabel='Cond. parameter')
    
    return fig, ax


def plot_cdf(parameters, labels, parameter, cluster_names=None, colors=None, 
             figsize=(5,5), parameter_names=None, legend_names=None,
             plot_prior=False):
    """Plot the class-conditional cdf for a single parameter.
    
    params:
        parameters [array(float)]: array of shape (n_parameters, n_simulations) 
                containing the parameter sets used for each model run
        labels [array(int)]: array of length n_simulations where each value 
                represents the cluster to which that model belongs
        parameter [int]: index of the parameter to plot, corresponding to a 
                column within `parameters`
        cluster_names [list(str)]: ordered list of cluster names corresponding
                to the label array, ie the 0th element of cluster_names is the
                name of the cluster where labels==0. Optional, default is 
                ['Cluster 0', 'Cluster 1', ...]
        colors [list]: colors to plot
        figsize [tuple(float)]: matplotlib figure size in inches. Optional, 
                default: (5,5)
        parameter_name [str]: name of the parameter as listed on the x-axis 
                label. Optional, defaults to 'Parameter #'
        legend_names [list(str)]: ordered list of names to display in the 
                legend. Optional, but must be a permuted version of 
                cluster_names.
                
        
    returns:
        fig: matplotlib figure handle
        ax: matplotlib axis handle
    """
    
    # Check input
    n_clusters = labels.max() + 1
    n_parameters = parameters.shape[1]

    if colors is None:
        colors = []
        cmap = matplotlib.cm.get_cmap('Set1')
        for i in range(n_clusters):
            colors.append(cmap(i))

    if cluster_names is None:
        cluster_names = ['Cluster %s' %i for i in range(n_clusters)]
    
    # If parameter is given as string, look for it in parameter_names
    if isinstance(parameter, str):
        if parameter_names is None:
            raise ValueError('Must provide list of parameter names if providing \
                parameter to plot as a string.')
        elif parameter not in parameter_names:
            raise ValueError('Could not find %s in parameter_names' % parameter)
        else:
            param_idx = parameter_names.index(parameter)
            
    elif isinstance(parameter, int):
        if parameter > (n_parameters - 1):
            raise ValueError('Parameter index passed (%s) is greater than the \
                             length of parameter_names' % parameter)
        else:
            param_idx = parameter
        
    if parameter_names is None:
        parameter_names = ['Parameter %s' %i for i in range(n_parameters)]
    
    # Get list of parameters (used for labeling with full col name)
    percentiles = np.arange(1, 100)

    fig, ax = plt.subplots(figsize=figsize, tight_layout=True)

    if plot_prior:
        ax.plot(np.percentile(parameters[:, param_idx], percentiles), 
                    percentiles, color='k', linestyle='--', label='$F$ ($X_i$)')

    for i in range(n_clusters):
        x = np.percentile(parameters[np.where(labels==i), param_idx], percentiles)
        ax.plot(x, percentiles, color=colors[i], 
                label='$F$ ($X_i$ | %s)' % cluster_names[i])

    ax.set(xlabel=parameter_names[param_idx], ylabel='% of simulations')

    # If legend_names is provided, sort the handles before adding legend
    if legend_names is None:
        ax.legend()
    else:
        handles, legend_labels = ax.get_legend_handles_labels()
        reordered_handles = []
        for name in legend_names:
            idx = legend_labels.index('$F$ ($X_i$ | %s)' % name)
            reordered_handles.append(handles[idx])

        ax.legend(reordered_handles, legend_names)

    return fig, ax


def plot_pdf(parameters, labels, parameter, cluster_names=None, colors=None,
             figsize=(5,5), parameter_names=None, legend_names=None,
             plot_prior=False, models_per_bin=5, **kwargs):
    """Plot the class-conditional pdf for a single parameter.

    params:
        parameters [array(float)]: array of shape (n_parameters, n_simulations)
                containing the parameter sets used for each model run
        labels [array(int)]: array of length n_simulations where each value
                represents the cluster to which that model belongs
        parameter [int]: index of the parameter to plot, corresponding to a
                column within `parameters`
        cluster_names [list(str)]: ordered list of cluster names corresponding
                to the label array, ie the 0th element of cluster_names is the
                name of the cluster where labels==0. Optional, default is
                ['Cluster 0', 'Cluster 1', ...]
        colors [list]: colors to plot
        figsize [tuple(float)]: matplotlib figure size in inches. Optional,
                default: (5,5)
        parameter_name [str]: name of the parameter as listed on the x-axis
                label. Optional, defaults to 'Parameter #'
        legend_names [list(str)]: ordered list of names to display in the
                legend. Optional, but must be a permuted version of
                cluster_names.
        models_per_bin [int]: this defines the spacing of the x-values in
                the pdf plot. A larger number means larger bins, which means
                lower resolution along the x-axis. Optional, default: 5
        **kwargs: keyword arguments to pass to scipy.stats.gaussian_kde


    returns:
        fig: matplotlib figure handle
        ax: matplotlib axis handle
    """

    # Check input
    n_clusters = labels.max() + 1
    n_parameters = parameters.shape[1]

    if colors is None:
        colors = []
        cmap = matplotlib.cm.get_cmap('Set1')
        for i in range(n_clusters):
            colors.append(cmap(i))

    if cluster_names is None:
        cluster_names = ['Cluster %s' %i for i in range(n_clusters)]

    # If parameter is given as string, look for it in parameter_names
    if isinstance(parameter, str):
        if parameter_names is None:
            raise ValueError('Must provide list of parameter names if providing \
                parameter to plot as a string.')
        elif parameter not in parameter_names:
            raise ValueError('Could not find %s in parameter_names' % parameter)
        else:
            param_idx = parameter_names.index(parameter)

    elif isinstance(parameter, int):
        if parameter > (n_parameters - 1):
            raise ValueError('Parameter index passed (%s) is greater than the \
                             length of parameter_names' % parameter)
        else:
            param_idx = parameter

    # Get list of parameters (used for labeling with full col name)
    if parameter_names is None:
        parameter_names = ['Parameter %s' %i for i in range(n_parameters)]

    fig, ax = plt.subplots(figsize=figsize, tight_layout=True, facecolor='white')

    if plot_prior:
        n = int(labels.shape[0]/models_per_bin)
        pmin = parameters[:, param_idx].min()
        pmax = parameters[:, param_idx].max()
        prange = pmax - pmin
        pmin -= prange*0.2
        pmax += prange*0.2
        x = np.linspace(pmin, pmax, n)
        kernel = gaussian_kde(parameters[:, param_idx].flatten(), **kwargs)
        pdf = kernel(x)
        ax.fill_between(x, 0, pdf, alpha=0.6, label='$f$ ($X_i$)', color='k', zorder=1)
        ax.plot(x, pdf, color='k', alpha=0.8, zorder=1)

    for i in range(n_clusters):
        n = int((labels == i).sum()/models_per_bin)
        label = '$f$ ($X_i$ | %s)' % cluster_names[i]
        pmin = parameters[np.where(labels==i), param_idx].min()
        pmax = parameters[np.where(labels==i), param_idx].max()
        prange = pmax - pmin
        pmin -= prange*0.2
        pmax += prange*0.2
        x = np.linspace(pmin, pmax, n)
        kernel = gaussian_kde(parameters[np.where(labels==i), param_idx].flatten(), **kwargs)
        pdf = kernel(x)

        ax.fill_between(x, 0, pdf, alpha=0.6, label=label, color=colors[i], zorder=1+((i+1)/n_clusters))
        ax.plot(x, pdf, color=colors[i], alpha=0.8, zorder=1.05+((i+1)/n_clusters))

    ax.set(xlabel=parameter_names[param_idx], ylabel='Density')

    # If legend_names is provided, sort the handles before adding legend
    if legend_names is None:
        ax.legend()
    else:
        handles, legend_labels = ax.get_legend_handles_labels()
        reordered_handles = []
        for name in legend_names:
            idx = legend_labels.index('$F$ ($X_i$ | %s)' % name)
            reordered_handles.append(handles[idx])

        ax.legend(reordered_handles, legend_names)

    return fig, ax



# -*- coding: utf-8 -*-

"""
Main pyDGSA module that provides the following functions:

dgsa 
dgsa_interactions
"""

import numpy as np
import pandas as pd
from .interact_util import interact_distance, interact_boot_distance
from tqdm.notebook import tqdm

def cluster_cdf_distance(prior_cdf, cluster_parameters, percentiles=None):
    """Helper function to calculate the L1-norm distance between the prior 
    distribution of each parameter and the class-conditional distribution 
    of each parameter. In other words, the distance between the cdf of a 
    parameter for all model runs and the cdf of a parameter for model runs 
    within an individual cluster.
    
    params:
        parameters [array(float)]: array of shape (n_parameters, n_simulations) 
                containing the parameter sets used for each model run
        cluster [array(int)]: array of indices corresponding to the models within 
                this cluster. Indices must match rows in parameters.
        
    returns:
        cluster_distance [array(float)]: array of shape (n_parameters,) containing 
                the distance between cdfs for each parameter
    """
    
    if percentiles is None:
        percentiles = np.arange(1,100)
        
    # Calculate the cdf of parameters in this individual cluster
    cluster_cdf = np.percentile(cluster_parameters, percentiles, axis=0)

    # Calculate the L1-norm between the cdf of parameters in this cluster
    # and the cdf of the entire population
    cluster_distance = np.sum(abs(cluster_cdf - prior_cdf), axis=0)

    return cluster_distance


def dgsa(parameters, labels, parameter_names=None, n_boots=3000, quantile=0.95, 
         output='mean', cluster_names=None, confidence=False):
    """Given a parameter set and clustered model responses, calculate the 
    normalized model sensitivity to each parameters, without interactions.
    
    params:
        parameters [array(float)]: array of shape (n_parameters, n_simulations) 
                containing the parameter sets used for each model run
        labels [array(int)]: array of length n_simulations where each value represents 
                the cluster to which that model belongs.
        parameter_names [list(str)]: list of parameter names. If not provided, will 
                default to ['param0', 'param1', ..., 'paramN']
        n_boots [int]: number of bootstrapped datasets to create. Default: 3000
        quantile [float]: quantile used to test the null hypothesis. Can specify 
                as a percentile [0-100] or quantile [0-1]. Default: 0.95
        confidence [bool]: whether to output confidence intervals on the standardized
                sensitivity measure, following Park et al. (2015). 
        output [str]: whether to return the mean standardized sensitivity across 
                clusters ('mean'), the max sensitivity across clusters ('max'), or 
                the standardized sensitivity for each cluster ('cluster_avg').
    
    returns:
        df [DataFrame]: pandas dataframe containing the normalized sensitivity of 
                each parameter. If output='mean' or output='max', indices are 
                parameter names and the single column is 'sensitivity'. If 
                output='cluster_avg', indices are the parameter names and each column 
                is an individual cluster.
    """    

    
    ### Check the input and convert to dict
    n_clusters = labels.max() + 1
    n_parameters = parameters.shape[1]
    
    # Generate default parameter names if not provided
    if parameter_names is None:
        parameter_names = ['param{}'.format(i) for i in range(n_parameters)]
    
    if output not in ['cluster_avg', 'mean', 'max']:
        raise ValueError("Requested output format must be 'cluster_avg', 'mean' or 'max'")

    # Convert label array to dict of cluster indices
    # Also check that there are at least 10 models per class [Fenwick et al., 2014]
    clusters = {}
    for nc in range(n_clusters):
        cluster = np.where(labels == nc)[0]
        clusters[nc] = cluster
        if len(cluster) < 10:
            print("Warning: \
                  Cluster {} contains {} models. \
                  Recommend >=10 models per cluster for optimal performance.".format(nc, 
                                                                        len(cluster)))
    
    ### Step 1. 
    # Caclulate distances between the prior cdfs and the cluster-conditional cdf

    # Calculate the prior cdf
    percentiles = np.arange(1,100)
    prior_cdf = np.percentile(parameters, percentiles, axis=0)

    # Iterate over each cluster 
    cluster_distances = np.zeros((n_clusters, n_parameters))
    for i, cluster in clusters.items():
        # Get the model parameters for this cluster
        cluster_parameters = parameters[cluster]
        cluster_distances[i, :] = cluster_cdf_distance(prior_cdf, cluster_parameters, 
                                                       percentiles=percentiles)

    ### Step 2. 
    # Now that we have the distances for each cluster, test for statistical 
    # significance using bootstrapping
    boot_distances = np.zeros((n_boots, n_clusters, n_parameters))

    # Loop through n_boots
    for nb in tqdm(range(n_boots)):
        # For each boot, generate random clusters the same size as our clusters
        # and calc cdf distance between those
        for i, cluster in clusters.items():
            
            # Remove extra cluster dim if provided (prevents shape mismatch error)
            cluster = cluster.squeeze()

            # Generate random indices of size equal to that of the current cluster
            boot_idx = np.random.choice(np.arange(parameters.shape[0]), 
                                        size=len(cluster), 
                                        replace=False)
            boot_parameters = parameters[boot_idx]

            # Calculate cdf distance for this random cluster
            boot_distances[nb,i,:] = cluster_cdf_distance(prior_cdf, 
                                                          boot_parameters, 
                                                          percentiles=percentiles)

    # Get the nth quantile of all boots
    boot_quantiles = np.quantile(boot_distances, quantile, axis=0)

    ### Step 3. 
    # Calculate the normalized distances (d/d_95)
    cluster_sensitivities = cluster_distances/boot_quantiles

    # Calculate the confidence interval, if requested
    if confidence:
        dist_upper = np.quantile(boot_distances, 1, axis=0)
        dist_lower = np.quantile(boot_distances, quantile - (1-quantile),
                                  axis=0)
        upper_bound = cluster_distances/dist_upper
        lower_bound = cluster_distances/dist_lower
        
    # Output as a dataframe
    if output == 'mean':
        columns = ['sensitivity']
        data = np.mean(cluster_sensitivities, axis=0)
        
        if confidence:
            conf = np.mean(lower_bound - upper_bound, axis=0)
            columns.append('confidence')
            data = np.column_stack((data, conf))

        df = pd.DataFrame(data, index=parameter_names, columns=columns)

    elif output == 'max':
        columns = ['sensitivity']
        data = np.max(cluster_sensitivities, axis=0)
        
        if confidence:
            conf = np.max(lower_bound - upper_bound, axis=0)
            columns.append('confidence')
            data = np.column_stack((data, conf))
            
        df = pd.DataFrame(data, index=parameter_names, columns=columns)
        
    elif output == 'cluster_avg':
        if cluster_names is None:
            cluster_names = ['Cluster ' + str(i) for i in range(n_clusters)]
        columns = cluster_names.copy()
        
        data = cluster_sensitivities.T
        
        if confidence:
            conf = lower_bound.T - upper_bound.T
            for i, cluster in enumerate(cluster_names):
                columns.insert(2*i + 1, cluster + '_conf')
                data = np.insert(data, 2*i + 1, conf[:, i], axis=1)
        
        df = pd.DataFrame(data, index=parameter_names, columns=columns)

    return df


def dgsa_interactions(parameters, labels, cond_parameters=None, n_bins=3, 
                      parameter_names=None, quantile=0.95, n_boots=3000,
                      output='mean', confidence=False, cluster_names=None):
    """Given a parameter set, clustered model responses, and a list of 
    conditional parameters, calculate the sensitivity of each two-way 
    parameter interaction.
    
    params:
        parameters [array(float)]: array of shape (n_parameters, n_simulations) 
                containing the parameter sets used for each model run
        labels [array(int)]: array of length n_simulations where each value 
                represents the cluster to which that model belongs.
        cond_parameters [list(str|int)]: list of conditional parameters for 
                which to calculate interactions. Can be either a list of indices
                corresponding to columns of parameters, or a list of names 
                corresponding to parameter_names. Default: all parameters
        n_bins [int]: number of bins in which to separate out each 
                conditional parameter
        parameter_names [list(str)]: ordered list of parameter names matching 
                column order within parameters. If not provided, will default 
                to ['param0', 'param1', ..., 'paramN']
        n_boots [int]: number of bootstrapped datasets to create. Default: 3000
        quantile [float]: quantile used to test the null hypothesis. Can specify 
                as a percentile [0-100] or quantile [0-1]. Default: 0.95
        output [str]: format in which to return sensitivities. 
                    -'mat' returns sensitivities in matrix form (main params 
                    as columns and conditional params as indices). 
                    -'mean' returns in single column form with each individual
                    interaction spelled out in the indices (e.g., 'x | y')
                    -'cluster_avg' returns the average sensitivity for each
                    each cluster, without bin-specific values
                    -'bin_avg' returns the average sensitivity for each bin,
                    without cluster-specific values
                    -'indiv' returns the sensitivity for each bin/cluster 
                    combination. No averaging is performed.
                    
    
    returns:
        df [DataFrame]: pandas dataframe containing the normalized sensitivity 
        of each two-way parameter interaction. Columns correspond to each 
        parameter in parameter_names and indices are the conditonal parameters
        in cond_parameters.
    """    

    # Get number of clusters, parameters, etc
    n_parameters = parameters.shape[1]
    n_clusters = labels.max() + 1
    
    # Generate default list of parameter names
    if parameter_names is None:
        parameter_names = ['param{}'.format(i) for i in range(n_parameters)]

    if cond_parameters is None:
        cond_parameters=parameter_names

    if output not in ['mat', 'mean', 'cluster_avg', 'bin_avg', 'indiv']:
        raise ValueError("Output format must be 'mean', 'mat', 'indiv', 'bin_avg', or 'cluster_avg'")

    # if cond_parameters is a list of str, assume those are cond parameter names
    if isinstance(cond_parameters[0], str):
        cond_parameter_idx = [parameter_names.index(p) for p in cond_parameters]
        cond_parameter_idx.sort()
    # else if it's a list of int, assume those are cond parameter indices
    elif isinstance(cond_parameters[0], int):
        cond_parameter_idx = sorted(cond_parameters)

    n_cond_parameters = len(cond_parameter_idx)
    cond_parameter_names = [parameter_names[i] for i in cond_parameter_idx]
    
    # Convert label array to dict of indices
    clusters = {}
    for nc in range(n_clusters):
        cluster = np.where(labels == nc)[0]
        clusters[nc] = cluster
        
    # Calculate the thresholds separating bins for each parameter
    # thresholds is of shape (n_bins, n_parameters)
    thresholds = np.quantile(parameters, [b/n_bins for b in range(1, n_bins)], 
                             axis=0)
    percentiles = np.arange(1,100)
    
    # NaN-filled arrays of L1 distances and alpha-quantile L1 distances
    param_interact = np.empty((n_cond_parameters, n_parameters-1, n_clusters, 
                               n_bins), dtype='float64')
    param_interact[:] = np.nan
    boot_interact = np.empty((n_cond_parameters, n_parameters-1, n_clusters, 
                              n_bins), dtype='float64')
    boot_interact[:] = np.nan
    
    # Loop through each conditional parameter and calculate 
    for i, cond_idx in enumerate(tqdm(cond_parameter_idx, 
                                      desc='Performing DGSA')):
        param_interact[i] = interact_distance(cond_idx, parameters, clusters, 
                                                thresholds, percentiles)
        boot_interact[i] = interact_boot_distance(cond_idx, parameters, clusters, 
                                                    thresholds, percentiles, 
                                                    n_boots=n_boots, alpha=quantile)
    
    # Calculate the normalized distances (d/d_95)
    normalized_interactions = param_interact/boot_interact

    # Average over each bin
    sensitivity_per_class = np.nanmean(normalized_interactions, axis=3)
    
    # Average over each cluster
    sensitivity_per_bin = np.nanmean(normalized_interactions, axis=2)
    
    # Average over each cluster and bin
    sensitivity_over_class = np.nanmean(sensitivity_per_class, axis=2)
    
    ### Choose how to output the results
    # Output mean sensitvitiy as df, with indices as interactions (e.g., 'x | y')
    if output == 'mean':
        df = pd.DataFrame(columns=['sensitivity'])
        
        for i, cond_param in enumerate(cond_parameter_names):
            main_params = [x for x in parameter_names if x != cond_param]
            
            for j, main_param in enumerate(main_params):
                interact_name = main_param + ' | ' + cond_param
                df.loc[interact_name, 'sensitivity'] = sensitivity_over_class[i, j]
        df.sort_values(by='sensitivity', ascending=False, inplace=True)
        
    # Output mean sensitivity in array form, with rows as cond params and
    # cols as main params
    elif output == 'mat':
        df = pd.DataFrame(columns=parameter_names)
        
        for i, cond_param in enumerate(cond_parameter_names):
            main_params = [x for x in parameter_names if x != cond_param]
            
            for j, main_param in enumerate(main_params):
                df.loc[cond_param, main_param] = sensitivity_over_class[i, j]
    
    # Output sensitivity of each cluster, indices as interactions, cols as clusters
    elif output == 'cluster_avg':
        if cluster_names is None:
            cluster_names = ['Cluster {}'.format(i) for i in range(n_clusters)]
        df = pd.DataFrame(columns=cluster_names)
        
        for i, cond_param in enumerate(cond_parameter_names):
            main_params = [x for x in parameter_names if x != cond_param]
            
            for j, main_param in enumerate(main_params):
                interact_name = main_param + ' | ' + cond_param
                df.loc[interact_name, :] = sensitivity_per_class[i, j, :]
    
    # Output sensitivity of each cluster/bin combination with indices as 
    # interactions (e.g., 'x | y') and cols as a cluster/bin combination     
    elif output == 'bin_avg':
        bin_names = ['Bin {}'.format(i) for i in range(n_bins)]
        df = pd.DataFrame(columns=bin_names)
        
        for i, cond_param in enumerate(cond_parameter_names):
            main_params = [x for x in parameter_names if x != cond_param]

            for j, main_param in enumerate(main_params):
                interact_name = main_param + ' | ' + cond_param
                df.loc[interact_name, :] = sensitivity_per_bin[i, j, :]

    # Output sensitivity of each cluster/bin combination with indices as 
    # interactions (e.g., 'x | y') and cols as a cluster/bin combination     
    elif output == 'indiv':
        if cluster_names is None:
            cluster_names = ['Cluster {}'.format(i) for i in range(n_clusters)]
        bin_names = ['Bin {}'.format(i) for i in range(n_bins)]
        columns = pd.MultiIndex.from_product([cluster_names, bin_names])
        df = pd.DataFrame(columns=columns)
        idx = pd.IndexSlice
        
        for i, cond_param in enumerate(cond_parameter_names):
            main_params = [x for x in parameter_names if x != cond_param]

            for j, main_param in enumerate(main_params):
                interact_name = main_param + ' | ' + cond_param
                
                for k, cluster in enumerate(cluster_names):
                    df.loc[interact_name, idx[cluster, :]] = \
                        normalized_interactions[i, j, k, :]
  
    return df
    
